﻿using System;
using System.Collections.Generic;
using MVR.FileManagementSecure;
using UnityEngine;

namespace SceneNinja {

    [Serializable]
    public class PlayerData {
        [SerializeField] public int Index;
        [SerializeField] public string Id;
        [SerializeField] public string Label;
        [SerializeField] public bool RightSide;

        [SerializeField] public string AppearancePath;

        [SerializeField] public string PluginsPath;

        [SerializeField] public bool SuppressClothing;

        [SerializeField] public bool SuppressScale;

        [SerializeField] public bool MergePlugins;

        [SerializeField] public bool IgnorePluginVersion;


        private JSONStorableBool _suppressClothing;
        private JSONStorableBool _suppressScale;
        private JSONStorableBool _mergePlugins;
        private JSONStorableBool _ignorePluginVersion;

        private JSONStorableString _appearancePath;
        private JSONStorableString _pluginsPath;

        private UIDynamicTextField _appearanceLabel;
        private UIDynamicTextField _pluginLabel;

        private SceneNinja _ninja;

        public PlayerData() {
            SuperController.LogMessage("Initializing from JSON...");
        }

        public void LoadVars() {
            _suppressClothing = new JSONStorableBool(Id + "SuppressClothing", SuppressClothing, BoolUpdated);
            _suppressScale = new JSONStorableBool(Id + "SuppressScale", SuppressScale, BoolUpdated);
            _mergePlugins = new JSONStorableBool(Id + "MergePlugins", MergePlugins, BoolUpdated);
            _ignorePluginVersion = new JSONStorableBool(Id + "IgnorePluginVersion", IgnorePluginVersion, BoolUpdated);
            _appearancePath = new JSONStorableString(Id + "ApperancePath", AppearancePath, StringUpdated);
            _pluginsPath = new JSONStorableString(Id + "PluginsPath", PluginsPath, StringUpdated);
            if (Id.Contains("M")) {
                Label = "Male " + Id.Replace("M", "");
                RightSide = true;
            } else {
                Label = "Female " + Id.Replace("F", "");
                RightSide = false;
            }
        }

        public PlayerData(string id) {
            Id = id;
            _suppressClothing = new JSONStorableBool(Id + "SuppressClothing", true, BoolUpdated);
            _suppressScale = new JSONStorableBool(Id + "SuppressScale", true, BoolUpdated);
            _mergePlugins = new JSONStorableBool(Id + "MergePlugins", true, BoolUpdated);
            _ignorePluginVersion = new JSONStorableBool(Id + "IgnorePluginVersion", true, BoolUpdated);
            _appearancePath = new JSONStorableString(Id + "ApperancePath", "", StringUpdated);
            _pluginsPath = new JSONStorableString(Id + "PluginsPath", "", StringUpdated);
            if (id.Contains("M")) {
                Index = 1;
                if (id.Contains("2")) Index = 3;
                Label = "Male " + id.Replace("M", "");
                RightSide = true;
            } else {
                Index = 0;
                if (id.Contains("2")) Index = 2;
                Label = "Female " + id.Replace("F", "");
                RightSide = false;
            }

            AppearancePath = "";
            PluginsPath = "";
            SuppressClothing = true;
            SuppressScale = true;
            MergePlugins = true;
        }

        private void StringUpdated(JSONStorableString js) {
            SuperController.LogMessage($"String updated: {js.name}" + js.val);
            var name = js.name;
            if (name.Contains("Appearance")) {
                AppearancePath = js.val;
            } else {
                PluginsPath = js.val;
            }
            _ninja.UpdatePlayer(this);
        }

        private void BoolUpdated(JSONStorableBool jb) {
            var name = jb.name;
            SuperController.LogMessage($"Bool updated: {name}" + jb.val);
            if (name.Contains("Clothing")) SuppressClothing = jb.val;
            if (name.Contains("Scale")) SuppressScale = jb.val;
            if (name.Contains("Merge")) MergePlugins = jb.val;
            if (name.Contains("Ignore")) IgnorePluginVersion = jb.val;
            _ninja.UpdatePlayer(this);
        }

        public void Initialize(SceneNinja ninja) {
            _ninja = ninja;
            ninja.RegisterBool(_suppressClothing);
            ninja.RegisterBool(_suppressScale);
            ninja.RegisterBool(_mergePlugins);
            ninja.RegisterBool(_ignorePluginVersion);
            ninja.RegisterString(_appearancePath);
            ninja.RegisterString(_pluginsPath);
            ninja.CreateUILabel(Label, RightSide);
            ninja.CreateUIButton(Id + " Appearance", RightSide, SelectAppearanceFile);
            ninja.CreateUIButton("Clear Appearance", RightSide, ClearAppearanceFile);
            _appearanceLabel = ninja.CreateUITextField(_appearancePath, RightSide, 40);
            ninja.CreateUIToggle(_suppressClothing, RightSide, "Suppress Clothing Load");
            ninja.CreateUIToggle(_suppressScale, RightSide, "Suppress Scale Load");
            ninja.CreateUIButton(Id + " Plugins", RightSide, SelectPluginPreset);
            ninja.CreateUIButton("Clear Plugins", RightSide, ClearPluginsFile);
            _pluginLabel = ninja.CreateUITextField(_pluginsPath, RightSide, 40);
            ninja.CreateUIToggle(_mergePlugins, RightSide, "Merge Load Plugins");
            ninja.CreateUIToggle(_ignorePluginVersion, RightSide, "Ignore Plugin Versions");
        }

        private void ClearPluginsFile() {
            PluginFileSelected("");
        }

        private void ClearAppearanceFile() {
            AppearanceFileSelected("");
        }

        private void SelectPluginPreset() {
            var shortcuts = FileManagerSecure.GetShortCutsForDirectory("Custom\\Atom\\Person\\Plugins", false, false, true);
            SuperController.singleton.GetMediaPathDialog(PluginFileSelected, "vap", "Custom\\Atom\\Person\\Plugins", false,
                true, false, null, false, shortcuts);
        }

        private void SelectAppearanceFile() {
            List<ShortCut> shortcuts =
                FileManagerSecure.GetShortCutsForDirectory("Custom\\Atom\\Person\\Appearance", false, false, true);
            SuperController.singleton.GetMediaPathDialog(AppearanceFileSelected, "vap",
                "Custom\\Atom\\Person\\Appearance", false, true, false, null, false, shortcuts);
        }

        private void PluginFileSelected(string fileName) {
            var file = "";
            if (fileName != "" && !fileName.Contains("Appearance")) {
                PluginsPath = fileName;
                var parts = fileName.Split('/');
                file = parts[parts.Length - 1];
                SuperController.LogMessage($"File selected for {Label}: " + file);
            }
            _pluginsPath.val = fileName;
            _pluginLabel.text = file;
        }

        private void AppearanceFileSelected(string fileName) {
            var file = "";
            if (fileName != "" && !fileName.Contains("Plugins")) {
                AppearancePath = fileName;
                var parts = fileName.Split('/');
                file = parts[parts.Length - 1];
                SuperController.LogMessage($"Appearance selected for {Label}: " + file);
            }
            _appearancePath.val = fileName;
            _appearanceLabel.text = file;
        }
    }
}
