using System;
using System.Collections.Generic;
using System.Linq;
using System.Text.RegularExpressions;
using JetBrains.Annotations;
using MeshVR;
using MVR.FileManagementSecure;
using SimpleJSON;
using UnityEngine;
using UnityEngine.Events;
using UnityEngine.UI;
using JSONClass = SimpleJSON.JSONClass;

namespace SceneNinja {

    public class SceneNinja : MVRScript
    {
        private const string _settingsDir = "Saves/PluginData/Venkman/SceneNinja";
        private const string _settingsPath = "Saves/PluginData/Venkman/SceneNinja/SceneNinjaDefaults.json";
        private PlayerData[] _players;

        public override void Init() {
            _players = new PlayerData[4];
            // Example storable; you can also create string, float and action JSON storables
            var players = new List<PlayerData>();

            var foo = FileManagerSecure.FileExists(_settingsPath);
            if (foo) {
                SuperController.LogMessage("We have a config!");
                var raw = FileManagerSecure.ReadAllText(_settingsPath);
                var text = raw.Split(char.Parse("\n"));
                foreach (var line in text) {
                    try {
                        var player = JsonUtility.FromJson<PlayerData>(line);
                        player.LoadVars();
                        SuperController.LogMessage("Player: " + player.Label);
                        players.Add(player);
                    } catch (Exception) {
                        //
                    }
                }
            } else {
                SuperController.LogMessage("Creating default config.");
                var p1 = new PlayerData("F1");
                var p2 = new PlayerData("M1");
                var p3 = new PlayerData("F2");
                var p4 = new PlayerData("M2");
                players.Add(p1);
                players.Add(p2);
                players.Add(p3);
                players.Add(p4);
                _players = players.ToArray();
                SaveConfig();
            }

            _players = players.ToArray();
            foreach (var atom in _players) {
                atom.Initialize(this);
            }
            OnSceneLoaded();
        }


        // called zero
        void Awake()
        {
            SuperController.LogMessage($"SceneNinja: Awake");
        }

        // called first
        void OnEnable()
        {
            SuperController.LogMessage($"SceneNinja: Enable");

            SuperController.singleton.onSceneLoadedHandlers += OnSceneLoaded;
        }

        // called second
        void OnSceneLoaded()
        {
            SuperController.LogMessage("Scene loaded!");
            var atoms = SuperController.singleton.GetAtoms();
            var guys = new List<Atom>();
            var gals = new List<Atom>();
            foreach (var atom in atoms.Where(atom => atom.type == "Person")) {
                if (IsMale(atom)) {
                    SuperController.LogMessage("Guy: " + atom.name);
                    guys.Add(atom);
                } else {
                    SuperController.LogMessage("Gal: " + atom.name);
                    gals.Add(atom);
                }
            }

            var closestMale = GetClosestPerson(guys);
            var closestFemale = GetClosestPerson(gals);

            if (closestMale != null) {
                foreach (var person in guys) {
                    var playerData = _players[3];
                    if (person.name == closestMale.name) {
                        playerData = _players[1];
                    }
                    ApplyAppearance(person, playerData);
                    ApplyPlugins(person, playerData);
                }
            }
            if (closestFemale != null) {
                foreach (var person in gals) {
                    var playerData = _players[2];
                    if (person.name == closestFemale.name) {
                        playerData = _players[0];
                    }
                    ApplyAppearance(person, playerData);
                    ApplyPlugins(person, playerData);
                }
            }

        }

        private void ApplyPlugins(Atom target, PlayerData playerData) {
            var presetFilePath = playerData.PluginsPath;
            var normalPath = presetFilePath == "" ? "" : SuperController.singleton.NormalizePath(presetFilePath);
            var atom = SuperController.singleton.GetAtomByUid(target.name);
            manager = atom.GetStorableByID("PluginManager") as MVRPluginManager;
            var existingSettings = new JSONClass();
            existingSettings.Add("setUnlistedParamsToDefault", "true");
            var exStorables = new JSONArray();
            var existing = new Dictionary<string, JSONNode>();
            var indices = new List<int>();

            foreach (var st in atom.GetStorableIDs()) {
                if (st.Contains("plugin#")) {
                    exStorables.Add(st,atom.GetStorableByID(st).GetJSON(true, true, true));
                }
            }
            if (manager != null) {
                var managerJson = manager.GetJSON();
                exStorables.Add("PluginManager", managerJson);
                var plugins = manager.GetJSON()["plugins"];
                foreach (KeyValuePair<string, JSONNode> f in plugins.AsObject) {
                    existing[f.Key] = f.Value;
                    var index = f.Key.Replace("plugin#", "");
                    var idx = int.Parse(index);
                    indices.Add(idx);
                }
            }

            existingSettings.Add("storables", exStorables);

            if (presetFilePath != "" && FileManagerSecure.FileExists(presetFilePath)) {
                var presetPackageName = "";
                if (presetFilePath.Contains(":")) presetPackageName = presetFilePath.Substring(0, presetFilePath.IndexOf(':'));

                SuperController.LogMessage($"Looping plugins in preset ({atom.name}): " + normalPath);
                var data = FileManagerSecure.ReadAllText(normalPath);
                var presetJson = JSON.Parse(data);

                if (presetPackageName != "")
                {
                    var presetJsonString = presetJson.ToString();
                    if (presetJsonString.Contains("SELF:"))
                    {
                        presetJsonString = presetJsonString.Replace("SELF:", presetPackageName + ":");
                        presetJson = JSON.Parse(presetJsonString);
                    }
                }

                var mappings = new Dictionary<string, JSONNode>();
                var plugs = presetJson["storables"][0]["plugins"];
                foreach (KeyValuePair<string, JSONNode> storableJson in plugs.AsObject) {
                    mappings[storableJson.Key] = storableJson.Value;
                }

                var ignoreVersion = playerData.IgnorePluginVersion;
                var newIndexes = new Dictionary<string, string>();
                if (playerData.MergePlugins) {
                    SuperController.LogMessage("Merging plugin data?");
                    foreach (var map in mappings) {
                        var val = map.Value.ToString();
                        foreach (var ex in existing) {
                            var exVal = ex.Value.ToString();
                            if (ignoreVersion) {
                                var valPart = val.Split(char.Parse(":"));
                                var exPart = exVal.Split(char.Parse(":"));
                                var valParts = valPart[0].Split(char.Parse("."));
                                var exParts = exPart[0].Split(char.Parse("."));
                                valParts[valParts.Length - 1] = "latest";
                                exParts[exParts.Length - 1] = "latest";
                                exPart[0] = string.Join(".", exParts);
                                valPart[0] = string.Join(".", valParts);
                                var vval = string.Join(":", valPart);
                                var vexVal = string.Join(":", exPart);
                                val = vval;
                                exVal = vexVal;
                            }
                            if (val == exVal && map.Key != ex.Key) {
                                newIndexes[map.Key] = ex.Key.Replace("\"", "");
                                SuperController.LogMessage($"Match for {val}, mapping {map.Key} to {ex.Key}.");
                                break;
                            }
                        }
                    }

                    foreach (var map in mappings) {
                        var index = map.Key;
                        if (newIndexes.ContainsKey(index)) continue;
                        if (existing.ContainsKey(index)) {
                            for (var i = 0; i < 1000; i++) {
                                if (!indices.Contains(i)) {
                                    indices.Add(i);
                                    newIndexes[index] = "plugin#" + i;
                                    break;
                                }
                            }
                        }
                    }
                }

                var newObj = new JSONClass();
                foreach (KeyValuePair<string, JSONNode> node in presetJson.AsObject) {
                    if (node.Key == "storables") {
                        var storables = node.Value.AsArray;
                        var newStorables = new JSONArray();
                        for (var i = 0; i < storables.Count; i++) {
                            var storable = storables[i].AsObject;
                            var id = storable["id"].Value;
                            id = id.Replace("\"", "");
                            SuperController.LogMessage("Checking storable " + id);
                            if (id.Contains("PluginManager")) {
                                SuperController.LogMessage("Got pluginmanager.");
                                var plugins = storable["plugins"];
                                var newPlugins = new JSONClass();
                                foreach (KeyValuePair<string, JSONNode> p in plugins.AsObject) {
                                    SuperController.LogMessage("Checking plugin " + p.Key);
                                    if (newIndexes.ContainsKey(p.Key)) {
                                        SuperController.LogMessage($"Swapping plug from {p.Key} to {newIndexes[p.Key]}" );
                                        newPlugins.Add(newIndexes[p.Key], p.Value);
                                    } else {
                                        SuperController.LogMessage($"Leaving plug alone {p.Key}");
                                        newPlugins.Add(p.Key, p.Value);
                                    }
                                }
                                storable["plugins"] = newPlugins;
                                newStorables.Add(storable);
                            } else {
                                var idx = id.Split(char.Parse("_"))[0];
                                SuperController.LogMessage("Checking index for " + idx);
                                if (newIndexes.ContainsKey(idx)) {
                                    SuperController.LogMessage($"Updating storable from {idx} to {newIndexes[idx]}");
                                    var newIdx = newIndexes[idx];
                                    var newId = storable["id"].ToString().Replace(idx, newIdx);
                                    newId = newId.Replace("\"", "");
                                    storable["id"] = newId;
                                } else {
                                    SuperController.LogMessage($"No index for {id}");
                                }
                                newStorables.Add(storable);
                            }
                        }

                        newObj.Add("storables", newStorables);
                    } else {
                        SuperController.LogMessage("Nothing to do, storing key: " + node.Key);
                        newObj.Add(node.Key, node.Value);
                    }
                }

                if (playerData.MergePlugins) {
                    SuperController.LogMessage("Merging plugin data, forreal.");
                    var mergedStorables = new JSONArray();
                    var existingStorables = existingSettings["storables"].AsArray;
                    var newStorables = newObj["storables"].AsArray;
                    var newAdded = new List<string>();
                    var existingPlugins = new List<string>();
                    var plugMan = new JSONClass();
                    SuperController.LogMessage("Looping existing storables...again.");
                    foreach (JSONClass exS in existingStorables) {
                        var exId = exS["id"].ToString().Replace("\"","");
                        SuperController.LogMessage("Checking " + exId);
                        if (exId != "PluginManager") {
                            var replaced = false;
                            foreach (JSONClass newS in newStorables) {
                                var nuId = newS["id"].ToString().Replace("\"","");
                                if (nuId == exId) {
                                    SuperController.LogMessage($"Replacing settings for {exId}");
                                    replaced = true;
                                    mergedStorables.Add(newS);
                                    newAdded.Add(nuId);
                                    break;
                                }
                            }

                            if (!replaced) {
                                mergedStorables.Add(exS);
                            }
                        } else {
                            plugMan = exS.AsObject;
                            var plugins = exS["plugins"];
                            existingPlugins.AddRange(plugins.AsObject.Keys);
                        }
                    }

                    foreach (JSONClass newS in newStorables) {
                        var nuId = newS["id"].ToString().Replace("\"","");
                        if (nuId != "PluginManager") {
                            if (!newAdded.Contains(newS["id"])) {
                                SuperController.LogMessage("Adding new storable: " + nuId);
                                mergedStorables.Add(newS);
                            }
                        } else {
                            var newPlugins = newS["plugins"];
                            var plugins = plugMan["plugins"];
                            foreach(KeyValuePair<string, JSONNode> np in newPlugins.AsObject) {
                                if (!existingPlugins.Contains(np.Key)) {
                                    SuperController.LogMessage("Adding new plugin entry: " + np.Key);
                                    plugins.Add(np.Key, np.Value);
                                }
                            }
                            SuperController.LogMessage("DONE.");
                            plugMan["plugins"] = plugins;
                            var reOrdered = new JSONArray();
                            reOrdered.Add(plugMan.AsObject);
                            foreach (JSONClass ms in mergedStorables) {
                                reOrdered.Add(ms);
                            }
                            mergedStorables = reOrdered;
                        }
                    }

                    newObj["storables"] = mergedStorables;
                    SuperController.LogMessage("Saving file.");
                    var newPath = normalPath.Replace(".vap", ".vab");
                    FileManagerSecure.WriteAllText(newPath, newObj.ToString());
                }

                var presetJs = atom.GetStorableByID("PluginPresets");
                atom.SetLastRestoredData(newObj, true, true);
                var pm = presetJs.GetComponentInChildren<PresetManager>();
                pm.LoadPresetFromJSON(newObj);
            }
        }

        private void ApplyAppearance(Atom atom, PlayerData playerData) {
            var tempPresetLockStore = new PresetLockStore();
            if (atom.type == "Person") tempPresetLockStore.StorePresetLocks(atom, true, playerData.SuppressClothing);

            var presetPackageName = "";
            var presetFilePath = playerData.AppearancePath;
            if (presetFilePath == "") return;

            if (presetFilePath.Contains(":")) presetPackageName = presetFilePath.Substring(0, presetFilePath.IndexOf(':'));

            var presetJson = (JSONClass)SuperController.singleton.LoadJSON(SuperController.singleton.NormalizePath(presetFilePath));
            var presetStorablesJson = (JSONArray)presetJson["storables"];

            var atomScale = atom.GetStorableByID("rescaleObject").GetFloatJSONParam("scale");
            if (playerData.SuppressScale) {
                foreach (JSONClass storableJson in presetStorablesJson)
                {
                    if (storableJson["id"].Value == "rescaleObject") storableJson["scale"].AsFloat = atomScale.val;

                }
            }
            if (presetPackageName != "")
            {
                var presetJsonString = presetJson.ToString();
                if (presetJsonString.Contains("SELF:"))
                {
                    presetJsonString = presetJsonString.Replace("SELF:", presetPackageName + ":");
                    presetJson = (JSONClass)JSON.Parse(presetJsonString);
                }
            }

            var presetJs = atom.GetStorableByID("AppearancePresets");
            var pm = presetJs.GetComponentInChildren<PresetManager>();
            atom.SetLastRestoredData(presetJson);
            pm.LoadPresetFromJSON(presetJson);
            if (atom.type == "Person" && playerData.SuppressClothing) tempPresetLockStore.RestorePresetLocks(atom);
        }

        // called third
        void Start()
        {
            SuperController.LogMessage($"SceneNinja: Start");
        }

        // called when the game is terminated
        void OnDisable()
        {
            SuperController.LogMessage($"SceneNinja: Disabled");
            SuperController.singleton.onSceneLoadedHandlers -=  OnSceneLoaded;
        }



        public void CreateUIButton(string label, bool rightSide = false, UnityAction callback = null,
            float height = 50, string helpText = "")
        {
            var button = CreateButton(label, rightSide);
            button.height = height;
            if (callback != null) button.button.onClick.AddListener(callback);
        }

        public UIDynamicTextField CreateUITextField(JSONStorableString label, bool rightSide = false,
            int height = 50, string helpText = "") {
            var field = CreateTextField(label, rightSide);
            var layout = field.GetComponent<LayoutElement>();
            layout.preferredHeight = layout.minHeight = height;
            field.height = height;
            return field;
        }

        public void CreateUIToggle(JSONStorableBool label, bool rightSide = false, string toggleLabel = "", UnityAction<bool> callback = null,
            float height = 50, string helpText = "") {
            var toggle = CreateToggle(label, rightSide);
            toggle.height = height;
            if (!string.IsNullOrEmpty(toggleLabel)) toggle.label = toggleLabel;
            if (callback != null) toggle.toggle.onValueChanged.AddListener(callback);
        }


        private static void disableScrollOnText(UIDynamicTextField target) {
            var targetSr = target.UItext.transform.parent.transform.parent.transform.parent.GetComponent<ScrollRect>();
            if( targetSr != null ) {
                targetSr.horizontal = false;
                targetSr.vertical = false;
            }
        }

        public void CreateUILabel(string descTitle, bool rightSide, int fieldHeight = 50, string helpText = "") {
            var staticDescString = new JSONStorableString(descTitle.Replace(" ", ""),descTitle);
            var staticDescStringField = CreateTextField(staticDescString, rightSide);
            staticDescStringField.backgroundColor = new Color(1f, 1f, 1f, 0f);
            var layout = staticDescStringField.GetComponent<LayoutElement>();
            layout.preferredHeight = layout.minHeight = fieldHeight;
            staticDescStringField.height = fieldHeight;
            disableScrollOnText(staticDescStringField);
        }

        public void UpdatePlayer(PlayerData playerData) {
            var i = playerData.Index;
            _players[i] = playerData;
            SaveConfig();
        }

        private void SaveConfig() {
            if (!FileManagerSecure.DirectoryExists(_settingsDir)) {
                FileManagerSecure.CreateDirectory(_settingsDir);
            }
            var lines = new List<string>();
            foreach (var player in _players) {
                lines.Add(JsonUtility.ToJson(player));
            }

            var json = "";
            foreach (var line in lines) {
                json += line + "\n";
            }

            SuperController.LogMessage("Saving JSON: " + json);
            FileManagerSecure.WriteAllText(_settingsPath, json);
        }

        private static bool IsMale(Atom atom)
        {
            var isMale = false;
            // If the person atom is not "On", then we cant determine their gender it seems as GetComponentInChildren<DAZCharacter> just returns null
            if (atom.on && (atom.containingSubScene == null || atom.containingSubScene.containingAtom.on) && atom.type == "Person") isMale = atom.GetComponentInChildren<DAZCharacter>().name.StartsWith("male") ||
                atom.GetComponentInChildren<DAZCharacter>().name.StartsWith("jarlee") ||
                atom.GetComponentInChildren<DAZCharacter>().name.StartsWith("jarjulian") ||
                atom.GetComponentInChildren<DAZCharacter>().name.StartsWith("Futa");
            return isMale;
        }

        [CanBeNull]
        private static Atom GetClosestPerson(List<Atom> atoms) {
            Atom closestPerson = null;
            var m1 = float.MaxValue;
            var isIndexed = true;
            var minIndex = int.MaxValue;
            var personIdx = new Dictionary<int, Atom>();
            foreach (var person in atoms) {
                if (person.hidden && !SuperController.singleton.showHiddenAtoms) continue;
                var d2 = float.MaxValue;
                foreach (var rb in person.rigidbodies) {
                    var rbPosition = rb.transform.position;
                    var playerHeadTransform = CameraTarget.centerTarget.transform;
                    var distance = Vector3.Distance(rbPosition, playerHeadTransform.position);
                    var resultString = Regex.Match(person.name, @"\d+").Value;
                    try {
                        var idx = int.Parse(resultString);
                        personIdx[idx] = person;
                    } catch (Exception) {
                        isIndexed = false;
                    }

                    if (distance < d2) {
                        d2 = distance;
                    }

                }

                if (d2 < m1) {
                    closestPerson = person;
                }
                SuperController.LogMessage($"Distances for {person.name} are {d2}/{m1}");
            }

            if (isIndexed) {
                SuperController.LogMessage("Indexed.");
                foreach (var pi in personIdx) {
                    if (pi.Key < minIndex) {
                        minIndex = pi.Key;
                        closestPerson = pi.Value;
                    }
                }
            }
            if (closestPerson != null) SuperController.LogMessage("Closest person: " + closestPerson.name);
            return closestPerson;
        }
    }
    public class PresetLockStore
    {
        private bool _generalPresetLock;
        private bool _appPresetLock;
        private bool _posePresetLock;
        private bool _animationPresetLock;
        private bool _glutePhysPresetLock;
        private bool _breastPhysPresetLock;
        private bool _pluginPresetLock;
        private bool _skinPresetLock;
        private bool _morphPresetLock;
        private bool _hairPresetLock;
        private bool _clothingPresetLock;

        public void StorePresetLocks(Atom atom, bool clearAllLocks = false, bool lockClothingPreset = false)
        {
            var pmControlList = atom.presetManagerControls;
            foreach (var pmc in pmControlList)
            {
                switch (pmc.name) {
                    case "geometry":
                        _generalPresetLock = pmc.lockParams;
                        break;
                    case "AppearancePresets":
                        _appPresetLock = pmc.lockParams;
                        break;
                    case "PosePresets":
                        _posePresetLock = pmc.lockParams;
                        break;
                    case "AnimationPresets":
                        _animationPresetLock = pmc.lockParams;
                        break;
                    case "FemaleGlutePhysicsPresets":
                        _glutePhysPresetLock = pmc.lockParams;
                        break;
                    case "FemaleBreastPhysicsPresets":
                        _breastPhysPresetLock = pmc.lockParams;
                        break;
                    case "PluginPresets":
                        _pluginPresetLock = pmc.lockParams;
                        break;
                    case "SkinPresets":
                        _skinPresetLock = pmc.lockParams;
                        break;
                    case "MorphPresets":
                        _morphPresetLock = pmc.lockParams;
                        break;
                    case "HairPresets":
                        _hairPresetLock = pmc.lockParams;
                        break;
                    case "ClothingPresets":
                        _clothingPresetLock = pmc.lockParams;
                        break;
                }

                if (pmc.name == "ClothingPresets" && lockClothingPreset) pmc.lockParams = true;
                else if (clearAllLocks)
                {
                    pmc.lockParams = false;
                }
            }

        }
        public void RestorePresetLocks(Atom atom)
        {
            var pmControlList = atom.presetManagerControls;
            foreach (var pmc in pmControlList)
            {
                switch (pmc.name) {
                    case "geometry":
                        pmc.lockParams = _generalPresetLock;
                        break;
                    case "AppearancePresets":
                        pmc.lockParams = _appPresetLock;
                        break;
                    case "PosePresets":
                        pmc.lockParams = _posePresetLock;
                        break;
                    case "AnimationPresets":
                        pmc.lockParams = _animationPresetLock;
                        break;
                    case "FemaleGlutePhysicsPresets":
                        pmc.lockParams = _glutePhysPresetLock;
                        break;
                    case "FemaleBreastPhysicsPresets":
                        pmc.lockParams = _breastPhysPresetLock;
                        break;
                    case "PluginPresets":
                        pmc.lockParams = _pluginPresetLock;
                        break;
                    case "SkinPresets":
                        pmc.lockParams = _skinPresetLock;
                        break;
                    case "MorphPresets":
                        pmc.lockParams = _morphPresetLock;
                        break;
                    case "HairPresets":
                        pmc.lockParams = _hairPresetLock;
                        break;
                    case "ClothingPresets":
                        pmc.lockParams = _clothingPresetLock;
                        break;
                }
            }

        }

    }
}
